@@ -36,16 +36,16 @@
 #define NUM_TESTS       20      /* test this many lines from the log */
 #define MAX_LOG_ERRORS  20
 #define READ_BYTES      4096u
+#define MAX_BATCH_LINES 8192u   /* max number of lines to read per batch before a reflow */
 
 #define LINE_LEN        23
 #define ERROR_LEN       255
 #define REF_SITE_LEN    511     /* maximum length of a referring site */
 #define CACHE_STATUS_LEN 7
 
-#define SPEC_TOKN_SET   0x1
-#define SPEC_TOKN_NUL   0x2
-#define SPEC_TOKN_INV   0x3
-#define SPEC_SFMT_MIS   0x4
+#define SPEC_TOKN_NUL   0x1
+#define SPEC_TOKN_INV   0x2
+#define SPEC_SFMT_MIS   0x3
 
 #include "commons.h"
 #include "gslist.h"
